//////////////////////////////////////////////////////////////////////////////
///
///  @file utils.cpp
///  
///  @author ag  @date   Sep 5, 2017
//////////////////////////////////////////////////////////////////////////////

// Standard libraries and the CAPD library
#include <iostream>
#include "utils.h"
#include "capd/covrel/HSet2D.h"
#include "capd/dynsys/DynSysMap.h"
#include "capd/intervals/lib.h"
#include "capd/capdlib.h"
//#include <stdexcept>
//#include <sstream>
//#include "capd/vectalg/iobject.hpp"
//#include "capd/dynset/C0DoubletonSet.h"
//#include "capd/geomset/CenteredDoubletonSet.hpp"
//#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"
using namespace capd;
using namespace capd::alglib;
using namespace capd::matrixAlgorithms;
using namespace std;

typedef capd::covrel::HSet2D<DMatrix, IMatrix> HSet2D;

//////////////////////////////////////////////////////////////////////////////
///	Some auxiliary functions
//////////////////////////////////////////////////////////////////////////////

interval pi=interval::pi();

// modPi 
// modulo pi for intervals
interval modPi(interval x)					
{
	//interval pi=interval::pi();
	if(x>0.0) while(!(x<pi)){x=x-pi;}
	if(x<0.0) while(x<0.0){x=x+pi;}
	return x;
}

// mod2Pi
// modulo 2 pi for intervals
interval mod2Pi(interval x)					
{
	//interval pi=interval::pi();
	if(x>0.0) while(!(x<2*pi)){x=x-2*pi;}
	if(x<0.0) while(x<0.0){x=x+2*pi;}
	return x;
}

// modXZ
// modPi for first variable and mod2Pi for the third
// as the phase space for Hyperion model
IVector modXZ(IVector x)					
{											
	x[0]=modPi(x[0]);
	x[2]=mod2Pi(x[2]);
	return x;
}

// Tr
// trace for interval matrices 2x2
interval Tr(IMatrix Dp)						
{ 
  return Dp[0][0]+Dp[1][1];
}

// Det
// determinant for interval matrices 2x2
interval Det(IMatrix Dp)					
{
 
  return Dp[0][0]*Dp[1][1] - Dp[1][0]*Dp[0][1];
}

// Delta
// determinant of characteristic polynomial	
// for interval matrices 2x2
interval Delta(IMatrix Dp)																			
{											
  interval tr=Dp[0][0]+Dp[1][1];
  interval det=Dp[0][0]*Dp[1][1] - Dp[1][0]*Dp[0][1];
 
  return tr*tr - 4*det;
}

// cut
// cuts the third (or further) coordinate(s) of an interval vector
IVector cut(IVector x)						
{
	IVector y(2);
	y[0]=x[0];
	y[1]=x[1];
	return y;
}

// cuts an interval matrix to 2x2
IMatrix cut(IMatrix M)						
{
	IMatrix y(2,2);
	y[0][0]=M[0][0];
	y[0][1]=M[0][1];
	y[1][0]=M[1][0];
	y[1][1]=M[1][1];
	return y;
}

//expand
//expands an interval 2-vector to 3-vector with zero
IVector expand(IVector x)					
{											
	IVector y(3);
	y[0]=x[0];
	y[1]=x[1];
	return y;
}

//expands an interval 2x2 matrix to 3x3 with identity
IMatrix expand(IMatrix M)					
{											
	IMatrix y(3,3);
	y[0][0]=M[0][0];
	y[0][1]=M[0][1];
	y[1][0]=M[1][0];
	y[1][1]=M[1][1];
	y[0][2]=y[1][2]=y[2][0]=y[2][1]=interval(0);
	y[2][2]=interval(1);
	return y;
}

//srodek
//number in the middle of an interval
double srodek(const interval & x) 			 
{
	return x.leftBound()+0.5*(x.rightBound()-x.leftBound());	
}

/////////////////////////////////////////////////////////////////////////
/// SecMap class
/////////////////////////////////////////////////////////////////////////


// image, stores the derivative in DP
IVector SecMap::image(const IVector &x,IMatrix &DP) const  
{
	IVector X(3);
	X[0]=x[0];
	X[1]=x[1];
	
	C1Rect2Set Q(X);
	IMatrix DP3(3,3);
	
	IVector Y=(*P)(Q,DP3);
	DP3=P->computeDP(Y,DP3);
	
	IVector y(2);
	y[0]=modPi(Y[0]);
	y[1]=Y[1];
	
	for(int i=0;i<2;i++)
	{
		for(int j=0;j<2;j++) DP[i][j]=DP3[i][j];
	}
	return y;
}

// derivative only
IMatrix SecMap::derivative(const IVector &x) const		
{
	IMatrix DP(2,2);
	image(x,DP);
	return DP;
}

// image only, no derivative calculation (faster)
IVector SecMap::image(const IVector &x) const		
{
	IVector X(3);
	X[0]=x[0];
	X[1]=x[1];
	
	C0Rect2Set Q(X);
	
	IVector Y=(*P)(Q);
	
	IVector y(2);
	y[0]=modPi(Y[0]);
	y[1]=Y[1];
	
	return y;
}

/////////////////////////////////////////////////////////////////////////
/// PoincNaOsiSym class
/////////////////////////////////////////////////////////////////////////

// image, stores the derivative in DP (single interval)
interval PoincNaOsiSym::image(const interval &x,interval &DP) const			
{
  //interval pi=interval::pi();							
  IVector X(2),Y(2);
  IMatrix D(2,2);
  X[0]=pi/2;											
  X[1]=x;													
  Y=(*P)(X,D);										 
  interval res=Y[0]-pi/2;						
  DP=D[0][1];									
  return res;									
}

// image only, no derivative calculation (faster)
interval  PoincNaOsiSym::image(const interval &x) const			
{
  //interval pi=interval::pi();
  IVector X(2),Y(2);
  X[0]=pi/2;
  X[1]=x;
  Y=(*P)(X);

  interval res=Y[0]-pi/2;
  return res;
}

// derivative only
interval  PoincNaOsiSym::derivative(const interval &x) const			
{
  //interval pi=interval::pi();
  IVector X(2);
  IMatrix D(2,2);
  X[0]=pi/2;
  X[1]=x;
  (*P)(X,D);

  return D[0][1];
}

//////////////////////////////////////////////////////////////////////////////
///	Covering auxiliary functions
//////////////////////////////////////////////////////////////////////////////

// lewyBrzeg
// Left face of an interval vector at some coordinate 
IVector lewyBrzeg(const IVector &x, const int coord){
	IVector lewy=x;
	lewy[coord]=(x[coord]).leftBound();
	return lewy;
}

// prawyBrzeg
// Right face of an interval vector at some coordinate 
IVector prawyBrzeg(const IVector &x, const int coord){
	IVector prawy=x;
	prawy[coord]=(x[coord]).rightBound();
	return prawy;
}

// obrazPodzielony
// grids the interval vector N1 and calculates the image via Poincare map,
// in some affine coordinates (M1,p1) of N1 and image (M2, p2)
IVector obrazPodzielony(		
IPoincareMap &P,
const IVector &N1, 
const IMatrix &M1, 
const IVector &p1, 
const IMatrix &M2, 
const IVector &p2,
int podzialH,
int podzialV)
{											
	IVector kosteczka(3);	
	C0Rect2Set R0(p1,M1,midVector(N1));					// R0 = P + Mv * middle(N1) 
	IMatrix M2Inv=gaussInverseMatrix(M2);			
	interval rt;
	IVector obraz_kosteczki=P(R0,p2,M2Inv,rt);			// image of the middle point
	for(int i = 0; i<podzialH;i++)
	{
		for(int j=0;j<podzialV;j++)
		{
			kosteczka[0]=N1[0].left()+(interval(i)+interval(0.,1.))*(N1[0].right()-N1[0].left())/interval(podzialH);
			kosteczka[1]=N1[1].left()+(interval(j)+interval(0.,1.))*(N1[1].right()-N1[1].left())/interval(podzialV);
			C0Rect2Set R(p1,M1,kosteczka);					//R = P + Mv * b 
			obraz_kosteczki = intervalHull(obraz_kosteczki,P(R,p2,M2Inv,rt));	// image MvInv2*(P(Mv2*b2+p2)-p2) unified with previous images
		}
	}
return obraz_kosteczki;
}

// czyNakrywa
// checks if interval vector obrazX1 lies properly on X2:
// left side to the left of X2, right side to the right of X2,
// and intersection lies beween the top and bottom of X2 (spanned across)
bool czyNakrywa(const IVector &obrazX1, const IVector &X2) {
		IVector przeciecie = intersection(obrazX1,X2);
	bool czyPrzechodzi =(((lewyBrzeg(X2,1))[1]<przeciecie[1]) and ((prawyBrzeg(X2,1))[1]>przeciecie[1]));		// across?
	bool czyNaLewo =((lewyBrzeg(obrazX1,0))[0]<X2[0]);															// left side to the left?
	bool czyNaPrawo= ((prawyBrzeg(obrazX1,0))[0]>X2[0]);														// right side to the right?
		if(!czyPrzechodzi) {
			cout << "Not spanned across." << endl;
		}
		if(!czyNaLewo) {			
			cout << "Not on the left." << endl;
		}
		if(!czyNaPrawo) {
			cout << "Not on the right." << endl;
		}
		return (czyPrzechodzi and czyNaLewo and czyNaPrawo); 
	}

//////////////////////////////////////////////////////////////////////////////
///	mojUklad structure
//////////////////////////////////////////////////////////////////////////////

// looks for a stationary point on (x=pi/2) line with the starting point z0
// stores eingenvector matrix in Mv
IVector mojUklad::pStNaPi2(const interval &z0, IMatrix &Mv)
{
	//---------------------------------------------------------------------------
	// Poinc is the self-map of the {x=pi/2} line:
	// y -> P(pi/2, y, 0)[0] - pi/2
	// its zeroes are the stationary points on the {x=pi/2} line
	SecMap Poin(P);						// P 2-dim
	PoincNaOsiSym Poinc(Poin);			
	//---------------------------------------------------------------------------
	interval z, Z, DP, Pz, zPop;
	IVector p(3);		
	z= z0;	
	int a=1;								// loop numerator
	double delta = pow(10.0,-12);			// size of the iterated interval
	//---------------------------------------------------------------------------
        try{
			do{								// Preliminary check if the iteration is 'stable'
				Pz=Poinc(z,DP);
				zPop = z;
				z=z-Pz/DP;
				z=z.mid();				
				a++;
			} while(not(subset(z,zPop)) && a<20);

			// if z is a good candidate, then we try a proper interval Newton method:
			Z=z;
			Z=Z+delta*interval(-1.0,1.0);
           
			interval N=z-Poinc(z)/Poinc[Z];		// interval Newton

			if(subsetInterior(N,Z))					// is N contained in the primary neighbourhood?
				{
					//cout << "\nSuccess for starting point " << z0 << endl;
					p[0]=2.0*atan(1.0);				// = pi/2
					p[1]=N;							// p is the small interval vector containing stationary point
					IVector Y(2);
					Y[0]=2.0*atan(1.0);
					Y[1]=N;
					IMatrix U=Poin[Y];				// U = DP(p)
					if(Tr(U)>2)						// if p is hyperbolic
					{
						interval l1 = (Tr(U)+sqrt(Delta(U)))*0.5;		// eigenvalues
						interval l2 = (Tr(U)-sqrt(Delta(U)))*0.5;
						IVector v1(2);									// eigenvectors
						v1[0]=1;
						v1[1]=(l1-U[0][0])/U[0][1];
						IVector v2(2);
						v2[0]=1;
						v2[1]=(l2-U[0][0])/U[0][1];
						IVector v1N(2);
						v1N=v1/sqrt(power(v1[0],2.0)+power(v1[1],2.0));	// normalize
						IVector v2N(2);
						v2N=v2/sqrt(power(v2[0],2.0)+power(v2[1],2.0));							
						Mv[0][0]=v1N[0];								// store vectors in Mv
						Mv[0][1]=v2N[0];
						Mv[1][0]=v1N[1];
						Mv[1][1]=v2N[1];
						Mv[0][2]=Mv[1][2]=Mv[2][0]=Mv[2][1]=interval(0);	
						Mv[2][2]=interval(1);		
					}
				}
			else
				{
					cout << "\nFor the starting point " << z0 << " no success." << endl;
				}
			} catch(exception& e)
			{
				cout << "\nFor the starting point " << z0 << " exception caught: " << e.what() << endl;
            }
    return p;
}

// checks covering relation for the h-set of the form 
// p1 + M1*N1 to the h-set p2 + M2*N2 via Hyperion Poincare map
bool mojUklad::covers(
const IVector &N1, 
const IMatrix &M1, 
const IVector &p1, 
const IVector &N2, 
const IMatrix &M2, 
const IVector &p2){	
	C0Rect2Set X1(p1,M1,N1);		//X1 = p1 + M1 * N1 
	IMatrix M2Inv=gaussInverseMatrix(M2);	
	interval rt;	
	//---------------- correction by the number of periods on the z-coordinate  -------------------------------------------
	C0Rect2Set P1(p1);
	int liczba_okresow = round((srodek(P(P1)[0])-srodek(p1[0]))/(4.0*atan(1.0)));	// number of periods
	IVector pi00(3);
	pi00[0]=interval::pi();
	IVector obraz =cut(P(X1,p2+liczba_okresow*pi00,M2Inv,rt));						// corrected image
	return czyNakrywa(obraz,cut(N2));												// does the image cover N2?
}

// checks covering relation for the h-set of the form 
// p1 + M1*N1 to the h-set p2 + M2*N2 via Hyperion Poincare map
// with griding podzialH x podzialV
bool mojUklad::covers(
const IVector &N1, 
const IMatrix &M1, 
const IVector &p1, 
const IVector &N2, 
const IMatrix &M2, 
const IVector &p2,
int podzialH,
int podzialV){	
	IVector kosteczka(3);								// small cube
	C0Rect2Set R0(p1,M1,midVector(N1));					// R0 = p1 + M1 * middle point(N1) 	
	IMatrix M2Inv=gaussInverseMatrix(M2);	
	interval rt;
	//---------------- correction by the number of periods on the z-coordinate  -------------------------------------------
	C0Rect2Set P1(p1);
	int liczba_okresow = round((srodek(P(P1)[0])-srodek(p1[0]))/(4.0*atan(1.0)));	// number of periods
	IVector pi00(3);
	pi00[0]=interval::pi();
	IVector obraz_kosteczki=P(R0,p2+liczba_okresow*pi00,M2Inv,rt);			// image of the middle point, corrected by number of periods
	//---------------- gridding  -------------------------------------------
	for(int i = 0; i<podzialH;i++)											
	{
		for(int j=0;j<podzialV;j++)
		{
			kosteczka[0]=N1[0].left()+(interval(i)+interval(0.,1.))*(N1[0].right()-N1[0].left())/interval(podzialH);
			kosteczka[1]=N1[1].left()+(interval(j)+interval(0.,1.))*(N1[1].right()-N1[1].left())/interval(podzialV);
			C0Rect2Set R(p1,M1,kosteczka);															//R = p1 + M1 * small cube 
			obraz_kosteczki = intervalHull(obraz_kosteczki,P(R,p2+liczba_okresow*pi00,M2Inv,rt));	// image MvInv2*(P(Mv2*b2+p2)-p2) unified with previous images
		}
	}
	//cout << "N2 cube = " << N2 << endl;
	//cout << "gridded image = " << obraz_kosteczki << endl;
	return czyNakrywa(cut(obraz_kosteczki),cut(N2));
}

